import contextvars


current_koil = contextvars.ContextVar("current_koil", default=None)
current_loop = contextvars.ContextVar("current_loop", default=None)
current_taskclass = contextvars.ContextVar("current_taskclass", default=None)
current_cancel_event = contextvars.ContextVar("current_loop", default=None)



import asyncio
from contextlib import contextmanager
import os
import sys
import threading
from typing import Type
from koil.vars import *
import time


class KoilTask:
    def __init__(
        self, future=None, loop=None, *args, log_errors=True, **kwargs
    ) -> None:
        super().__init__()
        self.future = future
        self.log_errors = log_errors
        self.loop = loop

        self.task = None

    async def wrapped_future(self, future):
        try:
            return await future
        except Exception as e:
            raise e

    def run(self):
        self.task = self.loop.create_task(self.wrapped_future(self.future))
        return self

    async def acancel(self):
        try:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError as e:
                print("Cancelled KoilTask")
        except Exception as e:
            print("Koil Task Cancellation failed")

    def cancel(self):
        return asyncio.run_coroutine_threadsafe(self.acancel(), self.loop).result()

    def result(self):
        return asyncio.run_coroutine_threadsafe(self.acancel(), self.loop).result()


@contextmanager
def _selector_policy():
    original_policy = asyncio.get_event_loop_policy()
    try:
        if (
            sys.version_info >= (3, 8)
            and os.name == "nt"
            and hasattr(asyncio, "WindowsSelectorEventLoopPolicy")
        ):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        yield
    finally:
        asyncio.set_event_loop_policy(original_policy)


def run_threaded_event_loop(loop):
    asyncio.set_event_loop(loop)
    try:
        loop.run_forever()
    finally:
        try:
            # mimic asyncio.run() behavior
            # cancel unexhausted async generators
            tasks = asyncio.all_tasks(loop)
            for task in tasks:
                task.cancel()

            async def gather():
                await asyncio.gather(*tasks, return_exceptions=True)

            loop.run_until_complete(gather())
            for task in tasks:
                if task.cancelled():
                    continue
                if task.exception() is not None:
                    loop.call_exception_handler(
                        {
                            "message": "unhandled exception during loop shutdown",
                            "exception": task.exception(),
                            "task": task,
                        }
                    )
            if hasattr(loop, "shutdown_asyncgens"):
                loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            loop.close()


def get_threaded_loop(name="KoilLoop"):
    """Create or return the default Koil IO loop
    The loop will be running on a separate thread.
    """
    with _selector_policy():
        newloop = asyncio.new_event_loop()

    th = threading.Thread(target=run_threaded_event_loop, args=(newloop,), name=name)
    th.daemon = True
    th.start()

    return newloop


class Koil:
    def __init__(
        self, name="Koiloop", grace_period=None, task_class: Type[KoilTask] = KoilTask
    ) -> None:
        self.it = "it"
        self.name = name
        self.task_class = task_class

    def __enter__(self):
        self.loop = get_threaded_loop(self.name)
        current_loop.set(self.loop)
        current_taskclass.set(self.task_class)

    async def aclose(self):
        loop = asyncio.get_event_loop()
        print("Causing loop to stop")
        loop.stop()
        print("Loop Stopped")

    def __exit__(self, exc_type, exc_val, exc_tb):
        asyncio.run_coroutine_threadsafe(self.aclose(), self.loop)

        iterations = 0

        while self.loop.is_running():
            time.sleep(0.001)
            iterations += 1
            if iterations == 100:
                print(
                    "Shutting Down takes longer than expected. Probably we are having loose Threads? Keyboard interrupt?"
                )

        current_loop.set(None)
        current_taskclass.set(self.task_class)



import asyncio
from koil.vars import *
import time


def unkoil_gen(iterator, *args, timeout=None, **kwargs):
    loop = current_loop.get()
    cancel_event = current_cancel_event.get()

    if loop.is_closed():
        raise RuntimeError("Loop is not running")
    try:
        loop0 = asyncio.events.get_running_loop()
        if loop0 is loop:
            raise NotImplementedError("Calling sync() from within a running loop")
    except RuntimeError:
        pass

    ait = iterator(*args, **kwargs).__aiter__()
    res = [False, False]

    async def next_on_ait():
        try:
            try:
                obj = await ait.__anext__()
                return [False, obj]
            except StopAsyncIteration:
                return [False, obj]
        except asyncio.CancelledError as e:
            return [False, e]

    while True:
        res = asyncio.run_coroutine_threadsafe(next_on_ait(), loop=loop)
        while not res.done():
            if cancel_event and cancel_event.is_set():
                raise Exception("Task was cancelled")

            time.sleep(0.01)
        done, obj = res.result()
        if done:
            if obj:
                raise obj
            break
        yield obj


def unkoil(coro, *args, timeout=None, as_task=False, ensure_koiled=True, **kwargs):
    loop = current_loop.get()
    cancel_event = current_cancel_event.get()

    if loop:
        try:
            if loop.is_closed():
                raise RuntimeError("Loop is not running")
            try:
                loop0 = asyncio.events.get_running_loop()
                if loop0 is loop:
                    raise NotImplementedError(
                        "Calling sync() from within a running loop, async that bitch"
                    )
            except RuntimeError:
                pass

            if as_task:
                taskclass = current_task_class.get()
                assert taskclass is not None, "No task class set"
                return taskclass(coro(*args, **kwargs), loop).run()

            co_future = asyncio.run_coroutine_threadsafe(coro(*args, **kwargs), loop)
            while not co_future.done():
                time.sleep(0.01)
                if cancel_event and cancel_event.is_set():
                    raise Exception("Task was cancelled")

            return co_future.result()
        except KeyboardInterrupt:
            print("Grace period triggered?")
            raise
    else:
        if ensure_koiled:
            raise RuntimeError("No loop set and ensure_koiled was set to True")

        if as_task:
            raise RuntimeError(
                """No loop is running. That means you cannot have this run as a task. Try providing a loop by entering a Koil() context.\n

                If you spawned a thread, in an Excecutor. You can use check_cancel() if you want to check if the thread was cancelled. Do this periodically
                """
            )
        asyncio.run(future)
